/* ========================================================================
 * PlantUML : a free UML diagram generator
 * ========================================================================
 *
 * (C) Copyright 2009-2017, Arnaud Roques
 *
 * Project Info:  http://plantuml.com
 * 
 * If you like this project or if you find it useful, you can support us at:
 * 
 * http://plantuml.com/patreon (only 1$ per month!)
 * http://plantuml.com/paypal
 * 
 * This file is part of PlantUML.
 *
 * PlantUML is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PlantUML distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 *
 *
 * Original Author:  Arnaud Roques
 * 
 *
 */
package net.sourceforge.plantuml.donors;

import java.awt.geom.Dimension2D;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.StringTokenizer;

import net.sourceforge.plantuml.AbstractPSystem;
import net.sourceforge.plantuml.BackSlash;
import net.sourceforge.plantuml.FileFormatOption;
import net.sourceforge.plantuml.code.AsciiEncoder;
import net.sourceforge.plantuml.code.CompressionBrotli;
import net.sourceforge.plantuml.code.StringCompressorNone;
import net.sourceforge.plantuml.code.Transcoder;
import net.sourceforge.plantuml.code.TranscoderImpl;
import net.sourceforge.plantuml.core.DiagramDescription;
import net.sourceforge.plantuml.core.ImageData;
import net.sourceforge.plantuml.graphic.GraphicStrings;
import net.sourceforge.plantuml.graphic.HtmlColorUtils;
import net.sourceforge.plantuml.graphic.StringBounder;
import net.sourceforge.plantuml.graphic.TextBlock;
import net.sourceforge.plantuml.graphic.UDrawable;
import net.sourceforge.plantuml.svek.TextBlockBackcolored;
import net.sourceforge.plantuml.ugraphic.ColorMapperIdentity;
import net.sourceforge.plantuml.ugraphic.ImageBuilder;
import net.sourceforge.plantuml.ugraphic.UGraphic;
import net.sourceforge.plantuml.ugraphic.UImage;
import net.sourceforge.plantuml.ugraphic.UTranslate;
import net.sourceforge.plantuml.version.PSystemVersion;

public class PSystemDonors extends AbstractPSystem {

	public static final String DONORS = "6nC809m7nctnXhJewjSHZX8nZQRAWw08ArguWCYs4bz-N7zS5gS5n0MmUhAGwulEPXEoPaowUSvodb4K"
			+ "owS1wwKEszpi1x5KjlyoSRkk0xPKSoN4WrWl_Vbxzy1H88lRb4SO1pRVmiRSDeVT4vKDkjbzWHPG_Mtq"
			+ "icMVChT_HZoYHrz2DcMtRQS98eNV77TGoLAC_OWKNTtO0dIyeuve7pJ7-8Bwrt_K8hLepQOXxdi0iOqK"
			+ "SitRFcQFW-oVm429GgDEq82P0wc7_2A3lpyX0nvgrM57E8ffhBFwpXC9I9YWGWh-aIVLSmi9Zp8uK6XS"
			+ "7GZe_1X60FQgg1BEjcu3ShHWYwm60TsuNd1Bp1VJKrmU6YuZ9pGiQ0LKeSX5q0M-IfkvHBcFNM737zoT"
			+ "yPRrGLvbu97qPA6ljUuCcE_b4vceRc_gMmlfN7ATurzd0pnhoCpPRaQBHpR7RFsOIw7vGW7oIw5JOy1Y"
			+ "FB-uRXOI81XBwe8ovi3nwnG1ayvIjV2BMF8QzuykirXLn6yjMAYN-El6feH32wn4N0tuXfR3atM-KBnh"
			+ "g7yVSZLrV0kkSOYvjq3FdZRdiz_BRcbcb1wWrI4qGbT6c2lcZYfJB8O835FwFU0cIUwa-jlaKNrRFiJU"
			+ "zYSXl7dX0IbjXj3VLLVOYXnMGHP4EjZ50Z1P9Tg5Gkj6tE5HLtGHKEwXC7VzTPFkOTF6n7NGEwvb2sDG"
			+ "NwOJv1Tk5tD9Asdea6KpTrXxWr6ACxXnNZ_sifnL0EkUtasBKBp8WB-7ESDLUWt9ZoDjYLjHXyaijmNF"
			+ "wcBFuNRovw4JVpZcPr776ySVFFuSCXRLz_QffMuHvgx1yBSrP5bx1egom6AuRXkenepxWFCPoP5AZb2e"
			+ "XJCT2ZP9GczCNZiA5kNMJgo-Rjc-D2r1zfwhkq1tyiFo84qh-SGy2Zlx7SCpb4UH_UAyCxDkwlkGOcS1"
			+ "UMEOu2uLp352YchOd466p3dqj5idd_dI5WlM4TKcFEnG-9Y9bemRA6tbXXgYUzoWCf6gNDM9InWJ8O3l"
			+ "EFq27ZNh86UBbbQC0xYb90IeWfBR_c5iL6YxQnhh4hrksxQFPUEl_ynL5Zd5_1Ah1aXr49N_OWXl9XO5"
			+ "U71t-5xOK1y0DZB3EWm0";

	@Override
	final protected ImageData exportDiagramNow(OutputStream os, int num, FileFormatOption fileFormat, long seed)
			throws IOException {
		final UDrawable result = getGraphicStrings();
		final ImageBuilder imageBuilder = new ImageBuilder(new ColorMapperIdentity(), 1.0, HtmlColorUtils.WHITE,
				getMetadata(), null, 0, 0, null, false);
		imageBuilder.setUDrawable(result);
		return imageBuilder.writeImageTOBEMOVED(fileFormat, seed, os);
	}

	private UDrawable getGraphicStrings() throws IOException {
		final List<TextBlock> cols = getCols(getDonors(), 6, 5);
		return new UDrawable() {
			public void drawU(UGraphic ug) {
				final TextBlockBackcolored header = GraphicStrings.createBlackOnWhite(Arrays
						.asList("<b>Special thanks to our sponsors and donors !"));
				header.drawU(ug);
				final StringBounder stringBounder = ug.getStringBounder();
				ug = ug.apply(new UTranslate(0, header.calculateDimension(stringBounder).getHeight()));
				double x = 0;
				double lastX = 0;
				double y = 0;
				for (TextBlock tb : cols) {
					final Dimension2D dim = tb.calculateDimension(stringBounder);
					tb.drawU(ug.apply(new UTranslate(x, 0)));
					lastX = x;
					x += dim.getWidth() + 10;
					y = Math.max(y, dim.getHeight());
				}
				final UImage logo = new UImage(PSystemVersion.getPlantumlImage());
				ug.apply(new UTranslate(lastX, y - logo.getHeight())).draw(logo);
			}
		};
	}

	public static List<TextBlock> getCols(List<String> lines, final int nbCol, final int reserved) throws IOException {
		final List<TextBlock> result = new ArrayList<TextBlock>();
		final int maxLine = (lines.size() + (nbCol - 1) + reserved) / nbCol;
		for (int i = 0; i < lines.size(); i += maxLine) {
			final List<String> current = lines.subList(i, Math.min(lines.size(), i + maxLine));
			result.add(GraphicStrings.createBlackOnWhite(current));
		}
		return result;
	}

	private List<String> getDonors() throws IOException {
		final List<String> lines = new ArrayList<String>();
		final Transcoder t = new TranscoderImpl(new AsciiEncoder(), new StringCompressorNone(), new CompressionBrotli());
		final String s = t.decode(DONORS).replace('*', '.');
		final StringTokenizer st = new StringTokenizer(s, BackSlash.NEWLINE);
		while (st.hasMoreTokens()) {
			lines.add(st.nextToken());
		}
		return lines;
	}

	public DiagramDescription getDescription() {
		return new DiagramDescription("(Donors)");
	}

	public static PSystemDonors create() {
		return new PSystemDonors();
	}

}
